package ru.dbotthepony.kommons.random

import java.util.random.RandomGenerator

/**
 * Disclaimer written by author of this PRNG, Bob Jenkins:
 * > WARNING: Alpha code. I'm not happy with the amount of avalanche, or with it using a multiplication.
 * > This is indefinitely in alpha until I have time to explore this space properly.
 * > I may get a better idea and replace these and still call them the same name.
 * > In the meantime you are encouraged to time it and detect bias in it.
 * > It's faster than JFS64 on my home machine, but slower on my work machine.
 * > The difference seems to be that my work machine has a slow multiply.
 * > Chips with slow multiplies are very common.
 *
 * WOB2M (Wrangler Of Bits, 2 mixing variables, with Multiply) is a 64-bit noncryptographic pseudorandom number generator.
 * It uses 1 counter and 2 mixing variables. It is reversible, but does not have the ability to jump forward or back by arbitrary amounts.
 * All 1-mix 1-counter functions were too slow, so there will be no WOB or WOB1 or WOB1M. WOB2, WOB3, WOB3M are likely in the future.
 *
 * Its cycle length is guaranteed to be at least 2^64 values because it takes that long for the counter to wrap around.
 * Every seed produces a distinct sequence for at least 2^64 values, because every seed starts with the same counter value but different other state,
 * so it takes at least 2^64 values before the counter wraps around and the rest of the state gets a chance to collide with the initial state
 * of some other seed. The longest cycle is expected to be of length 2^191, containing roughly half of all possible states, and the average cycle
 * length is expected to be 2^190. Two arbitrary seeds will land on the same cycle with probability about 1/3, but that is OK because the sequences
 * won't overlap for at least 2^64 values.
 */
open class WOB2MRandom protected constructor(
	@JvmField
	protected var seed0: Long,
	@JvmField
	protected var seed1: Long,
	@JvmField
	protected var count: Long,
	marker: Nothing?
) : RandomGenerator {
	constructor(seed0: Long, seed1: Long) : this(seed0, seed1, Long.MIN_VALUE + 10, null) {
		for (i in 0 until 10)
			nextLong()
	}

	final override fun nextLong(): Long {
		val temp = seed0 + count++
		seed0 = seed1 + temp.rotateLeft(12)
		seed1 = (0x0581af43eb71d8b3 * temp) xor seed0.rotateLeft(28)
		return seed1
	}

	fun previousLong(): Long {
		val temp = 0x6cc3621b095c967b * (seed1 xor seed0.rotateLeft(28))
		seed1 = seed0 - temp.rotateLeft(12)
		seed0 = temp - --count
		return seed1
	}
}
